/**
 * @fileoverview This rule checks for category in sparrow track event calls
 * @author Sachin Fernandes
 */
'use strict';
//------------------------------------------------------------------------------
// Rule Definition
//------------------------------------------------------------------------------

const sparrowModuleName = '@cloudflare/util-sparrow';
const trackFunctionName = 'track';
const sparrowImportNamespace = 'sparrow';

let isSparrowTrackInScope = false;

function isValidType(argument, type) {
  return typeof argument !== 'undefined' && argument.type === type;
}

function hasKeyInObject(object, keyToFind) {
  // we want to capture both {'category': 'value'} and {category: 'value'}
  // When wrapped in quotes the key becomes a Literal and doesn't have a name but has a value
  return object.properties.some(function(property) {
    // We check for property.key to avoid failures on spread syntax
    // i.e. spread syntax has no property.key
    return (
      property.key &&
      (property.key.name === keyToFind || property.key.value === keyToFind)
    );
  });
}

function isTrackFunction(callee) {
  return (
    // We need to handle both track() and sparrow.track()
    callee.name === trackFunctionName ||
    (callee.property &&
      callee.property.name === trackFunctionName &&
      callee.object.name === sparrowImportNamespace)
  );
}

/**
 * Returns the name of the module imported or re-exported.
 *
 * @param {ASTNode} node - A node to get.
 * @returns {string} the name of the module, or empty string if no name.
 */
function getValue(node) {
  if (node && node.source && node.source.value) {
    return node.source.value.trim();
  }

  return '';
}

// Checks if we've imported/required sparrow track function
// returns true or false
function checkIfTrackInScope(node) {
  // If we've already detected sparrow track in scope, just return true
  if (isSparrowTrackInScope) {
    return true;
  }
  // Handles track requires
  // i.e. const sparrow = require('@cloudflare/util-sparrow');
  if (node.callee && node.callee.name === 'require') {
    return node.arguments.some(
      argument => argument.value === sparrowModuleName
    );
  }

  // Handles track imports like
  // import {track} from '@cloudflare/util-sparrow';
  const value = getValue(node);
  if (value === sparrowModuleName) {
    return node.specifiers.some(
      specifier =>
        (specifier.type === 'ImportSpecifier' &&
          specifier.imported.name === trackFunctionName) ||
        specifier.type === 'ImportNamespaceSpecifier'
    );
  }
  return false;
}

module.exports = {
  meta: {
    docs: {
      description:
        'This rule checks for recommended best practices for sparrow',
      category: 'Possible Errors',
      recommended: false
    },
    fixable: null,
    schema: [],
    messages: {
      missing: 'Missing category in sparrow track call'
    }
  },

  create: function(context) {
    //----------------------------------------------------------------------
    // Public
    //----------------------------------------------------------------------

    isSparrowTrackInScope = false;
    const handleImportDeclaration = function(node) {
      isSparrowTrackInScope = checkIfTrackInScope(node);
    };

    // This function handles analyzing the actual call to the track funtion
    const handleCallExpression = function(node) {
      // Require statements are CallExpressions.
      // So we check if if the track funtion is in scope
      isSparrowTrackInScope = checkIfTrackInScope(node);

      if (isSparrowTrackInScope && isTrackFunction(node.callee)) {
        const arg1 = node.arguments[0];
        const arg2 = node.arguments[1];
        if (
          typeof arg1 !== 'undefined' &&
          isValidType(arg2, 'ObjectExpression')
        ) {
          if (!hasKeyInObject(arg2, 'category')) {
            context.report({
              node,
              loc: node.loc,
              messageId: 'missing'
            });
          }
        } else {
          context.report({
            node,
            loc: node.loc,
            messageId: 'missing'
          });
        }
      }
    };

    return {
      ImportDeclaration: handleImportDeclaration,
      CallExpression: handleCallExpression
    };
  }
};
